#define HASHTABLE_NAME		SymbolTable
#define HASHTABLE_ENTRY_NAME	Symbol
#define HASHTABLE_ENTRY_MEMBERS	int tally;
#define HASHTABLE_VERBOSE	1

#include "hashtable.c"

#include <stdio.h>
#include <assert.h>

struct Symbol *lookup(struct SymbolTable *t, char *key)
{
    unsigned len= strlen(key);
    struct Symbol *found= SymbolTable_find(t, key, len);
    if (!found) found= SymbolTable_insert(t, key, len);
    return found;
}

void test1(void)
{
    struct SymbolTable symbols;
    int i;
    char *words[]= {
	"You're", "obliged", "to", "pretend", "respect", "for", "people", "and", "institutions", "you", "think",
	"absurd", ".", "You", "live", "attached", "in", "a", "cowardly", "fashion", "to", "moral", "and", "social",
	"conventions", "you", "despise", ",", "condemn", ",", "and", "know", "lack", "all", "foundation", ".", "It", "is",
	"that", "permanent", "contradiction", "between", "your", "ideas", "and", "desires", "and", "all", "the",
	"dead", "formalities", "and", "vain", "pretenses", "of", "your", "civilization", "which", "makes", "you",
	"sad", ",", "troubled", "and", "unbalanced", ".", "That's", "the", "poisoned", "and", "mortal", "wound", "of",
	"the", "civilized", "world", ".",
		"The", "worship", "of", "money", "is", "the", "lowest", "of", "all", "human", "emotions", ",", "but", "it",
	"is", "shared", "not", "only", "by", "the", "bourgeoisie", "but", "also", "by", "the", "great", "majority",
	"of", "us", ":", "Little", "people", ",", "humble", "people", ",", "even", "those", "who", "are", "practically",
	"penniless", ".", "And", "I", ",", "with", "all", "my", "indignation", ",", "all", "my", "passion", "for",
	"destruction", ",", "I", ",", "too", ",", "am", "not", "free", "of", "it", ".", "I", "who", "am", "oppressed", "by",
	"wealth", ",", "who", "realise", "it", "to", "be", "the", "source", "of", "all", "misery", ",", "all", "my",
	"vices", "and", "hatred", ",", "all", "the", "bitterest", "humiliations", "that", "I", "have", "to", "suffer", ",",
	"all", "my", "impossible", "dreams", "and", "all", "the", "endless", "torment", "of", "my", "existence", ",",
	"still", ",", "all", "the", "time", ",", "as", "soon", "as", "I", "find", "myself", "in", "the", "presence", "of",
	"a", "rich", "person", ",", "I", "cannot", "help", "looking", "up", "to", "him", ",", "as", "some", "exceptional",
	"and", "splendid", "being", ",", "a", "kind", "of", "marvelous", "divinity", ",", "and", "in", "spite", "of",
	"myself", ",", "stronger", "than", "either", "my", "will", "or", "my", "reason", ",", "I", "feel", "rising", "from",
	"the", "very", "depths", "of", "my", "being", ",", "a", "sort", "of", "incense", "of", "admiration", "for",
	"this", "wealthy", "creature", ",", "who", "is", "all", "too", "often", "as", "stupid", "as", "he", "is",
	"pitiless", ".", "Isn't", "it", "crazy", "?", "And", "why", "...", "why", "?"
    };

    struct Symbol *nodes[sizeof(words) / sizeof(*words)];

    SymbolTable_initialise(&symbols, 0);

    for (i= 0;  i < sizeof(words) / sizeof(*words);  ++i) {
	int j;
	nodes[i]= lookup(&symbols, words[i]);
	nodes[i]->tally++;
	for (j= 0;  j <= i;  ++j) {
	    struct Symbol *addr= lookup(&symbols, words[j]);  (void)addr;
	    assert(nodes[j] == addr);
	}
    }
    for (i= 0;  i < symbols.hashtable_capacity;  ++i) {
	struct Symbol *s= symbols.hashtable_entries[i];
	if (s)
	    printf("%2i %-17s", s->tally, s->hashtable_entry_key);
    }
    printf("\n");
    SymbolTable_release(&symbols);
}

void test2(void)
{
    char *path= "/usr/share/dict/words";
    struct SymbolTable symbols;
    FILE *words= fopen(path, "r");
    if (!words) {
	fprintf(stderr, "I cannot open %s.  Edit me and try again.", path);
	return;
    }
    SymbolTable_initialise(&symbols, 0);
    for (;;) {
	char buf[256];
	if (!fgets(buf, sizeof(buf), words)) break;
	SymbolTable_insert(&symbols, buf, strlen(buf));
    }
    fclose(words);
    printf("interned %i symbols from %s\n", symbols.hashtable_tally, path);
    SymbolTable_release(&symbols);
}

int main(int argc, char **argv)
{
    int i;

    if (argc == 1) {
	test1();
	test2();
    }
    else {
	struct SymbolTable symbols;
	SymbolTable_initialise(&symbols, 8192);
	for (i= 1;  i < argc;  ++i) {
	    struct Symbol *n= lookup(&symbols, argv[i]);	    (void)n;
	    /* printf("%s %08x \t%p\n", argv[i], hash(argv[i], strlen(argv[i])), n); */
	}
	for (i= 1;  i < argc;  ++i) {
	    struct Symbol *n= lookup(&symbols, argv[i]);	    (void)n;
	    /* printf("%s %08x \t%p\n", argv[i], hash(argv[i], strlen(argv[i])), n); */
	}
	SymbolTable_release(&symbols);
    }

    return 0;
}
